/*
 * Copyright (C) 2006-2007 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file dccp_recv.c paraslash's dccp receiver */

/*
 * based on client.c of dccp-cs-0.01.tar.bz2,
 * (C) 2005 Ian McDonald <imcdnzl@gmail.com>
 */

#include "para.h"
#include "error.h"
#include "dccp.h"
#include "list.h"
#include "sched.h"
#include "recv.h"
#include "string.h"
#include "net.h"
#include "fd.h"

#include "dccp_recv.cmdline.h"

/* needed by getaddrinfo */
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>

/** the size of the output buffer */
#define DCCP_BUFSIZE 40960

/**
 * data specific to the dccp receiver
 *
 * \sa receiver receiver_node
 */
struct private_dccp_recv_data {
	/** the file descriptor for the dccp socket */
	int fd;
};


static void dccp_recv_close(struct receiver_node *rn)
{

	struct private_dccp_recv_data *pdd = rn->private_data;

	if (pdd && pdd->fd > 0)
		close(pdd->fd);
	free(rn->buf);
	rn->buf = NULL;
	free(rn->private_data);
	rn->private_data = NULL;
}


static int dccp_recv_open(struct receiver_node *rn)
{
	struct private_dccp_recv_data *pdd;
	struct dccp_recv_args_info *conf = rn->conf;
	int ret;
	struct addrinfo *ai;
	char *tmp;

	rn->buf = para_calloc(DCCP_BUFSIZE);
	rn->private_data = para_calloc(sizeof(struct private_dccp_recv_data));
	pdd = rn->private_data;
	ret = dccp_get_socket();
	if (ret < 0)
		goto err_out;
	pdd->fd = ret;

	tmp = make_message("%d", conf->port_arg);
	ret = getaddrinfo(conf->host_arg, tmp, NULL, &ai);
	free(tmp);
	if (ret) {
		ret = -E_ADDR_INFO;
		goto err_out;
	}
	ret = dccp_set_socket(pdd->fd);
	if (ret < 0)
		goto err_out;
	PARA_NOTICE_LOG("connecting to %s:%d\n", conf->host_arg, conf->port_arg);
	ret = connect(pdd->fd, ai->ai_addr, ai->ai_addrlen);
	freeaddrinfo(ai);
	if (ret < 0) {
		ret = -E_DCCP_CONNECT;
		goto err_out;
	}
	mark_fd_nonblock(pdd->fd);
	return 1;
err_out:
	dccp_recv_close(rn);
	return ret;
}

static void dccp_shutdown(void)
{
	; /* nothing to do */
}

static void *dccp_recv_parse_config(int argc, char **argv)
{
	struct dccp_recv_args_info *tmp = para_calloc(sizeof(struct dccp_recv_args_info));

	if (!dccp_recv_cmdline_parser(argc, argv, tmp))
		return tmp;
	free(tmp);
	return NULL;
}

static void dccp_recv_pre_select(struct sched *s, struct task *t)
{
	struct receiver_node *rn = t->private_data;
	struct private_dccp_recv_data *pdd = rn->private_data;

	t->ret = 1;
	para_fd_set(pdd->fd, &s->rfds, &s->max_fileno);
}

static void dccp_recv_post_select(struct sched *s, struct task *t)
{
	struct receiver_node *rn = t->private_data;
	struct private_dccp_recv_data *pdd = rn->private_data;

	t->ret = -E_DCCP_RECV_EOF;
	if (rn->output_eof && *rn->output_eof)
		goto out;
	t->ret = 1;
	if (!s->select_ret || !FD_ISSET(pdd->fd, &s->rfds))
		goto out; /* nothing to do */
	t->ret = -E_DCCP_OVERRUN;
	if (rn->loaded >= DCCP_BUFSIZE)
		goto out;
	t->ret = recv_bin_buffer(pdd->fd, rn->buf + rn->loaded,
		DCCP_BUFSIZE - rn->loaded);
	if (t->ret <= 0) {
		if (!t->ret)
			t->ret = -E_DCCP_RECV_EOF;
		goto out;
	}
	rn->loaded += t->ret;
	return;
out:
	if (t->ret < 0)
		rn->eof = 1;
}

/**
 * the init function of the dccp receiver
 *
 * \param r pointer to the receiver struct to initialize
 *
 * Initialize all function pointers of \a r
 */
void dccp_recv_init(struct receiver *r)
{
	r->shutdown = dccp_shutdown;
	r->open = dccp_recv_open;
	r->close = dccp_recv_close;
	r->pre_select = dccp_recv_pre_select;
	r->post_select = dccp_recv_post_select;
	r->parse_config = dccp_recv_parse_config;
}
