/*
 * Copyright (C) 1999-2007 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file mysql_selector.c para_server's mysql-based audio file selector */

/** \cond some internal constants */
#define MEDIUM_BLOB_SIZE 16777220 /*  (2**24 + 4) */
#define BLOB_SIZE 65539 /* (2**16 + 3) */
/** \endcond */
#include "server.cmdline.h"
#include "server.h"
#include "vss.h"
#include "afs_common.h"
#include <mysql/mysql.h>
#include <mysql/mysql_version.h>
#include <regex.h>
#include "error.h"
#include "net.h"
#include "string.h"
#include "user_list.h"
#include "mysql_selector_command_list.h"
#include "ipc.h"

/** pointer to the shared memory area */
extern struct misc_meta_data *mmd;

static void *mysql_ptr = NULL;
static int mysql_lock;

/**
 * contains name/replacement pairs used by s_a_r_list()
 *
 * \sa s_a_r()
 */
struct para_macro {
	/** the name of the macro */
	const char *name;
	/** the replacement text */
	const char *replacement;
};

static const struct para_macro mysql_macro_list[] = {
	{	.name = "IS_N_SET",
		.replacement = "(data.%s != '1')"
	}, {
		.name = "IS_SET",
		.replacement = "(data.%s = '1')"
	}, {
		.name = "PICID",
		.replacement = "%sdata.Pic_Id"
	}, {
		.name = "NAME_LIKE",
		.replacement = "(data.name like '%s')"
	}, {
		.name = "LASTPLAYED",
		.replacement = "%sFLOOR((UNIX_TIMESTAMP(now())"
			"-UNIX_TIMESTAMP(data.Lastplayed))/60)"
	}, {
		.name = "NUMPLAYED",
		.replacement = "%sdata.Numplayed"
	}, {
		.name = NULL,
	}
};

/**
 * Simple search and replace routine.
 *
 * \param src Source String.
 * \param macro_name The name of the macro.
 * \param replacement The replacement format string.
 *
 * In \p src, replace each occurence of \p macro_name(arg) by the string
 * determined by the \p replacement format string. \p replacement may (but
 * needs not) contain a single string conversion specifier (%s) which gets
 * replaced by \p arg.
 *
 * \return A string in which all matches in \p src are replaced, or \p NULL if
 * an error was encountered. Caller must free the result.
 *
 * \sa regcomp(3)
 */
__must_check __malloc static char *s_a_r(const char *src, const char* macro_name,
		const char *replacement)
{
	regex_t preg;
	size_t  nmatch = 1;
	regmatch_t pmatch[1];
	int eflags = 0;
	char *dest = NULL;
	const char *bufptr = src;

	if (!macro_name || !replacement || !src)
		return para_strdup(src);
	if (regcomp(&preg, macro_name, 0) != 0)
		return NULL;
	while (regexec(&preg,  bufptr, nmatch, pmatch, eflags)
			!= REG_NOMATCH) {
		char *tmp, *arg, *o_bracket, *c_bracket;

		o_bracket = strchr(bufptr + pmatch[0].rm_so, '(');
		c_bracket = o_bracket? strchr(o_bracket, ')') : NULL;
		if (!c_bracket)
			goto out;
		tmp = para_strdup(bufptr);
		tmp[pmatch[0].rm_so] = '\0';
		dest = para_strcat(dest, tmp);
		free(tmp);

		arg = para_strdup(o_bracket + 1);
		arg[c_bracket - o_bracket - 1] = '\0';
		tmp = make_message(replacement, arg);
		free(arg);
		dest = para_strcat(dest, tmp);
		free(tmp);
		bufptr = c_bracket;
		bufptr++;
	}
	dest = para_strcat(dest, bufptr);
//	PARA_DEBUG_LOG("%s: returning %s\n", __func__, dest);
out:
	regfree(&preg);
	return dest;
}

/**
 * replace a string according to a list of macros
 *
 * \param macro_list the array containing a macro/replacement pairs.
 * \param src the source string
 *
 * This function just calls s_a_r() for each element of \p macro_list.
 *
 * \return \p NULL if one of the underlying calls to \p s_a_r returned \p NULL.
 * Otherwise the completely expanded version of \p src is returned.
 */
__must_check __malloc static char *s_a_r_list(const struct para_macro *macro_list,
		char *src)
{
	const struct para_macro *mp = macro_list;
	char *ret = NULL, *tmp = para_strdup(src);

	while (mp->name) {
		ret = s_a_r(tmp, mp->name, mp->replacement);
		free(tmp);
		if (!ret) /* syntax error */
			return NULL;
		tmp = ret;
		mp++;
	}
	//PARA_DEBUG_LOG("%s: returning %s\n", __func__, dest);
	return ret;
}

static int lockless_real_query(const char *query)
{
	if (!mysql_ptr)
		return -E_NOTCONN;
	PARA_DEBUG_LOG("%s\n", query);
	if (mysql_real_query(mysql_ptr, query, strlen(query))) {
		PARA_ERROR_LOG("real_query error (%s)\n",
			mysql_error(mysql_ptr));
		return -E_QFAILED;
	}
	return 1;
}

static int real_query(const char *query)
{
	int ret;

	mutex_lock(mysql_lock);
	ret = lockless_real_query(query);
	mutex_unlock(mysql_lock);
	return ret;
}

/*
 * Use open connection given by mysql_ptr to query server. Returns a
 * result pointer on succes and NULL on errors
 */
static struct MYSQL_RES *get_result(const char *query)
{
	void *result = NULL;

	mutex_lock(mysql_lock);
	if (lockless_real_query(query) < 0)
		goto out;
	result = mysql_store_result(mysql_ptr);
	if (!result)
		PARA_ERROR_LOG("%s", "store_result error\n");
out:
	mutex_unlock(mysql_lock);
	return result;
}
/*
 * write input from fd to dynamically allocated char array,
 * but maximal max_size byte. Return size.
 */
static int fd2buf(int fd, char **buf_ptr, size_t max_size)
{
	const size_t chunk_size = 1024;
	size_t size = 2048;
	char *buf = para_malloc(size * sizeof(char)), *p = buf;
	int ret;

	while ((ret = recv_bin_buffer(fd, p, chunk_size)) > 0) {
		p += ret;
		if ((p - buf) + chunk_size >= size) {
			char *tmp;

			size *= 2;
			if (size > max_size) {
				ret = -E_TOOBIG;
				goto out;
			}
			tmp = para_realloc(buf, size);
			p = (p - buf) + tmp;
			buf = tmp;
		}
	}
	if (ret < 0)
		goto out;
	*buf_ptr = buf;
	ret = p - buf;
out:
	if (ret < 0 && buf)
		free(buf);
	return ret;
}

static char *escape_blob(const char* old, size_t size)
{
	char *new;

	if (!mysql_ptr)
		return NULL;
	new = para_malloc(2 * size * sizeof(char) + 1);
	mysql_real_escape_string(mysql_ptr, new, old, size);
	return new;
}

static char *escape_str(const char* old)
{
	return escape_blob(old, strlen(old));
}

static char *escaped_basename(const char *name)
{
	char *esc, *bn = para_basename(name);

	if (!bn)
		return NULL;
	esc = escape_str(bn);
	free(bn);
	return esc;
}

/*
 * new attribute
 */
int com_na(__a_unused int fd, int argc, char *argv[])
{
	char *q, *tmp;
	int ret;

	if (argc < 2)
		return -E_MYSQL_SYNTAX;
	tmp = escape_str(argv[1]);
	if (!tmp)
		return -E_ESCAPE;
	q = make_message("alter table data add %s char(1) "
		"not null default 0", tmp);
	free(tmp);
	ret = real_query(q);
	free(q);
	return ret;
}

/*
 * delete attribute
 */
int com_da(__a_unused int fd, int argc, char *argv[])
{
	char *q, *tmp;
	int ret;

	if (argc < 2)
		return -E_MYSQL_SYNTAX;
	tmp = escape_str(argv[1]);
	if (!tmp)
		return -E_ESCAPE;
	q = make_message("alter table data drop %s", tmp);
	free(tmp);
	ret = real_query(q);
	free(q);
	return ret;
}

/* stradd/pic_add */
static int com_stradd_picadd(int fd, int argc, char *argv[])
{
	char *blob = NULL, *esc_blob = NULL, *q = NULL, *tmp = NULL;
	const char *fmt, *del_fmt;
	int ret, stradd = strcmp(argv[0], "picadd");
	size_t size;

	if (argc < 2)
		return -E_MYSQL_SYNTAX;
	if (strlen(argv[1]) >= MAXLINE - 1)
		return -E_NAMETOOLONG;
	if (!mysql_ptr)
		return -E_NOTCONN;
	if (stradd) {
		size = BLOB_SIZE;
		fmt = "insert into streams (name, def) values ('%s','%s')";
		del_fmt="delete from streams where name='%s'";
	} else {
		size = MEDIUM_BLOB_SIZE;
		fmt = "insert into pics (name, pic) values ('%s','%s')";
		del_fmt="delete from pics where pic='%s'";
	}
	tmp = escape_str(argv[1]);
	if (!tmp)
		return -E_ESCAPE;
	q = make_message(del_fmt, tmp);
	free(tmp);
	ret = real_query(q);
	free(q);
	if (ret < 0)
		return ret;
	if ((ret = send_buffer(fd, AWAITING_DATA_MSG) < 0))
		return ret;
	if ((ret = fd2buf(fd, &blob, size)) < 0)
		return ret;
	size = ret;
	if (stradd)
		blob[size] = '\0';
	ret = -E_ESCAPE;
	esc_blob = escape_blob(blob, size);
	if (!esc_blob)
		goto out;
	tmp = escape_str(argv[1]);
	if (!tmp)
		goto out;
	q = make_message(fmt, tmp, esc_blob);
	ret = real_query(q);
out:
	free(blob);
	free(esc_blob);
	free(tmp);
	free(q);
	return ret;
}

/* stradd */
int com_stradd(int fd, int argc, char *argv[])
{
	return com_stradd_picadd(fd, argc, argv);
}

/* pic_add */
int com_picadd(int fd, int argc, char *argv[])
{
	return com_stradd_picadd(fd, argc, argv);
}

/*
 * print results to fd
 */
static int print_results(int fd, void *result,
		my_ulonglong top, my_ulonglong left,
		my_ulonglong bottom, my_ulonglong right)
{
	unsigned int i,j;
	int ret;
	MYSQL_ROW row;

	for (i = top; i <= bottom; i++) {
		row = mysql_fetch_row(result);
		if (!row || !row[0])
			return -E_NOROW;
		for (j = left; j <= right; j++) {
			ret = send_va_buffer(fd, j == left? "%s" : "\t%s",
					row[j]?  row[j] : "NULL");
			if (ret < 0)
				return ret;
		}
		ret = send_buffer(fd, "\n");
		if (ret < 0)
			return ret;
	}
	return 0;
}

/*
 * verbatim
 */
int com_verb(int fd, int argc, char *argv[])
{
	void *result = NULL;
	int ret;
	my_ulonglong num_rows, num_fields, top = 0, left = 0;
	char *tmp;

	if (argc < 2)
		return -E_MYSQL_SYNTAX;
	tmp = escape_str(argv[1]);
	if (!tmp)
		return -E_ESCAPE;
	result = get_result(tmp);
	free(tmp);
	if (!result)
		/* return success, because it's ok to have no results */
		return 1;
	num_fields = mysql_field_count(mysql_ptr);
	num_rows = mysql_num_rows(result);
	ret = 1;
	if (num_fields && num_rows)
		ret = print_results(fd, result, top, left, num_rows - 1,
			num_fields - 1);
	mysql_free_result(result);
	return ret;
}

/* returns NULL on errors or if there are no atts defined yet */
static void *get_all_attributes(void)
{
	void *result = get_result("desc data");
	unsigned int num_rows;

	if (!result)
		return NULL;
	num_rows = mysql_num_rows(result);
	if (num_rows < 5) {
		mysql_free_result(result);
		return NULL;
	}
	mysql_data_seek(result, (my_ulonglong)4); /* skip Lastplayed, Numplayed... */
	return result;
}

/*
 * list all attributes
 */
int com_laa(int fd, int argc, __a_unused char *argv[])
{
	void *result;
	int ret;
	my_ulonglong top = 0, left = 0, bottom, right = 0;

	if (argc != 1)
		return -E_MYSQL_SYNTAX;
	result = get_all_attributes();
	if (!result)
		return -E_NOATTS;
	bottom = mysql_num_rows(result);
	if (bottom < 5)
		return -E_MYSQL_SYNTAX;
	bottom -= 5;
	ret = print_results(fd, result, top, left, bottom, right);
	mysql_free_result(result);
	return ret;
}

/*
 * history
 */
int com_hist(int fd, int argc, char *argv[])
{
	int ret;
	void *result = NULL;
	char *q, *atts;
	my_ulonglong num_rows, top = 0, left = 0, right = 1;

	if (argc > 3)
		return -E_MYSQL_SYNTAX;
	if (argc > 1) {
		char *tmp = escape_str(argv[1]);
		if (!tmp)
			return -E_ESCAPE;
		atts = make_message("where %s = '1'", tmp);
		free(tmp);
	} else
		atts = para_strdup(NULL);

	q = make_message("select name, to_days(now()) - to_days(lastplayed) from "
		"data %s order by lastplayed", atts);
	free(atts);
	result = get_result(q);
	free(q);
	if (!result)
		return -E_NORESULT;
	num_rows = mysql_num_rows(result);
	ret = 1;
	if (num_rows)
		ret = print_results(fd, result, top, left, num_rows - 1, right);
	mysql_free_result(result);
	return ret;
}

/*
 * get last num audio files
 */
int com_last(int fd, int argc, char *argv[])
{
	void *result = NULL;
	char *q;
	int num, ret;
	my_ulonglong top = 0, left = 0, right = 0;

	if (argc < 2)
		num = 10;
	else
		num = atoi(argv[1]);
	if (!num)
		return -E_MYSQL_SYNTAX;
	q = make_message("select name from data order by lastplayed desc "
		"limit %u", num);
	result = get_result(q);
	free(q);
	if (!result)
		return -E_NORESULT;
	ret = print_results(fd, result, top, left, mysql_num_rows(result) - 1,
		right);
	mysql_free_result(result);
	return ret;
}

int com_mbox(int fd, int argc, char *argv[])
{
	void *result;
	MYSQL_ROW row;
	int ret;
	my_ulonglong num_rows, num_fields, top = 0, left = 0;
	char *query = para_strdup("select concat('From foo@localhost ', "
		"date_format(Lastplayed, '%a %b %e %T %Y'), "
		"'\nReceived: from\nTo: bar\n");

	ret = -E_NOATTS;
	result = get_all_attributes();
	if (!result)
		goto out;
	ret = -E_NOROW;
	while ((row = mysql_fetch_row(result))) {
		char *tmp;

		if (!row[0])
			goto out;
		tmp = make_message("%sX-Attribute-%s: ', %s, '\n", query,
			row[0], row[0]);
		free(query);
		query = tmp;
	}
	query = para_strcat(query,
		"From: a\n"
		"Subject: "
		"', name, '"
		"\n\n\n"
		"') from data"
	);
	if (argc >= 2) {
		char *esc = escape_str(argv[1]), *tmp;
		ret = -E_ESCAPE;
		if (!esc)
			goto out;
		tmp = make_message("%s where name LIKE '%s'", query, esc);
		free(esc);
		free(query);
		query = tmp;
	}
	mysql_free_result(result);
	ret = -E_NORESULT;
	result = get_result(query);
	if (!result)
		goto out;
	ret = -E_EMPTY_RESULT;
	num_fields = mysql_field_count(mysql_ptr);
	num_rows = mysql_num_rows(result);
	if (!num_fields || !num_rows)
		goto out;
	ret = print_results(fd, result, top, left, num_rows - 1,
		num_fields - 1);
out:
	free(query);
	if (result)
		mysql_free_result(result);
	return ret;
}

/*
 * get attributes by name. If verbose is not 0, this function returns a string
 * of the form 'att1="0",att2="1"'... which is used in com_cam() for
 * constructing a mysql update query. Otherwise the space-separated list of all
 * attributes which are set in the audio file given by name is returned.  Never
 * returns NULL in *NON VERBOSE* mode.
 */
static char *get_atts(char *name, int verbose)
{
	char *atts = NULL, *buf, *ebn;
	void *result = NULL, *result2 = NULL;
	MYSQL_ROW row, row2;
	int i;
	my_ulonglong num_fields, offset = 4; /* skip Lastplayed, Numplayed... */


	result2 = get_all_attributes();
	if (!result2)
		goto out;
	ebn = escaped_basename(name);
	if (!ebn)
		goto out;
	buf = make_message("select * from data where name='%s'", ebn);
	free(ebn);
	result = get_result(buf);
	free(buf);
	if (!result)
		goto out;
	num_fields = mysql_num_fields(result);
	if (num_fields < 5)
		goto out;
	mysql_data_seek(result2, offset);
	row = mysql_fetch_row(result);
	if (!row)
		goto out;
	for (i = 4; i < num_fields; i++) {
		int is_set = row[i] && !strcmp(row[i], "1");
		row2 = mysql_fetch_row(result2);
		if (!row2 || !row2[0])
			goto out;
		if (atts && (verbose || is_set))
			atts = para_strcat(atts, verbose? "," : " ");
		if (is_set || verbose)
			atts = para_strcat(atts, row2[0]);
		if (verbose)
			atts = para_strcat(atts, is_set? "=\"1\"" : "=\"0\"");
	}
out:
	if (result2)
		mysql_free_result(result2);
	if (result)
		mysql_free_result(result);
	if (!atts && !verbose)
		atts = para_strdup("(none)");
	return atts;
}

/* never returns NULL in verbose mode */
static char *get_meta(char *name, int verbose)
{
	MYSQL_ROW row;
	void *result = NULL;
	char *ebn, *q, *ret = NULL;
	const char *verbose_fmt =
		"select concat('lastplayed: ', "
		"(to_days(now()) - to_days(lastplayed)),"
		"' day(s). numplayed: ', numplayed, "
		"', pic: ', pic_id) "
		"from data where name = '%s'";
	/* is that really needed? */
	const char *fmt = "select concat('lastplayed=\\'', lastplayed, "
		"'\\', numplayed=\\'', numplayed, "
		"'\\', pic_id=\\'', pic_id, '\\'') "
		"from data where name = '%s'";

	if (!(ebn = escaped_basename(name)))
		goto out;
	q = make_message(verbose? verbose_fmt : fmt, ebn);
	free(ebn);
	result = get_result(q);
	free(q);
	if (!result)
		goto out;
	row = mysql_fetch_row(result);
	if (!row || !row[0])
		goto out;
	ret = para_strdup(row[0]);
out:
	if (result)
		mysql_free_result(result);
	if (!ret && verbose)
		ret = para_strdup("(not yet played)");
	return ret;
}

static char *get_dir(char *name)
{
	char *ret = NULL, *q, *ebn;
	void *result;
	MYSQL_ROW row;

	if (!(ebn = escaped_basename(name)))
		return NULL;
	q = make_message("select dir from dir where name = '%s'", ebn);
	free(ebn);
	result = get_result(q);
	free(q);
	if (!result)
		return NULL;
	row = mysql_fetch_row(result);
	if (row && row[0])
		ret = para_strdup(row[0]);
	mysql_free_result(result);
	return ret;
}

/* never returns NULL */
static char *get_current_stream(void)
{
	char *ret;
	MYSQL_ROW row;
	void *result = get_result("select def from streams where "
		"name = 'current_stream'");

	if (!result)
		goto err_out;
	row = mysql_fetch_row(result);
	if (!row || !row[0])
		goto err_out;
	ret = para_strdup(row[0]);
	mysql_free_result(result);
	return ret;
err_out:
	if (result)
		mysql_free_result(result);
	return para_strdup("(none)");
}

/*
 * Read stream definition of stream streamname and construct mysql
 * query. Return NULL on errors. If streamname is NULL, use current
 * stream. If that is also NULL, use query that selects everything.
 * If filename is NULL, query will list everything, otherwise only
 * the score of given file.
 */
static char *get_query(char *streamname, char *filename, int with_path)
{
	char *accept_opts = NULL, *deny_opts = NULL, *score = NULL;
	char *where_clause, *order, *query;
	char command[255] = ""; /* buffer for sscanf */
	void *result;
	MYSQL_ROW row;
	char *end, *tmp;
	char *select_clause = NULL;
	if (!streamname)
		tmp = get_current_stream();
	else {
		tmp = escape_str(streamname);
		if (!tmp)
			return NULL;
	}
	if (!strcmp(tmp, "(none)")) {
		free(tmp);
		if (filename) {
			char *ret, *ebn = escaped_basename(filename);
			if (!ebn)
				return NULL;
			ret = make_message("select to_days(now()) - "
				"to_days(lastplayed) from data "
				"where name = '%s'", ebn);
			free(ebn);
			return ret;
		}
		if (with_path)
			return make_message(
				"select concat(dir.dir, '/', dir.name) "
				"from data, dir where dir.name = data.name "
				"order by data.lastplayed"
			);
		return make_message(
			"select name from data where name is not NULL "
			"order by lastplayed"
		);
	}
	free(tmp);
	query = make_message("select def from streams where name = '%s'",
		streamname);
	result = get_result(query);
	free(query);
	query = NULL;
	if (!result)
		goto out;
	row = mysql_fetch_row(result);
	if (!row || !row[0])
		goto out;
	end = row[0];
	while (*end) {
		int n;
		char *arg, *line = end;

		if (!(end = strchr(line, '\n')))
			break;
		*end = '\0';
		end++;
		if (sscanf(line, "%200s%n", command, &n) < 1)
			continue;
		arg = line + n;
		if (!strcmp(command, "accept:")) {
			char *tmp2 = s_a_r_list(mysql_macro_list, arg);
			if (accept_opts)
				accept_opts = para_strcat(
					accept_opts, " or ");
			accept_opts = para_strcat(accept_opts, tmp2);
			free(tmp2);
			continue;
		}
		if (!strcmp(command, "deny:")) {
			char *tmp2 = s_a_r_list(mysql_macro_list, arg);
			if (deny_opts)
				deny_opts = para_strcat(deny_opts, " or ");
			deny_opts = para_strcat(deny_opts, tmp2);
			free(tmp2);
			continue;
		}
		if (!score && !strcmp(command, "score:"))
			score = s_a_r_list(mysql_macro_list, arg);
	}
	if (!score) {
		score = s_a_r_list(mysql_macro_list, conf.mysql_default_score_arg);
		if (!score)
			goto out;
	}
	if (filename) {
		char *ebn = escaped_basename(filename);
		if (!ebn)
			goto out;
		select_clause = make_message("select %s from data ", score);
		free(score);
		where_clause = make_message( "where name = '%s' ", ebn);
		free(ebn);
		order = para_strdup("");
		goto write_query;
	}
	select_clause = para_strdup(with_path?
		"select concat(dir.dir, '/', dir.name) from data, dir "
		"where dir.name = data.name "
		:
		"select name from data where name is not NULL");
	order = make_message("order by -(%s)", score);
	free(score);
	if (accept_opts && deny_opts) {
		where_clause = make_message("and ((%s) and not (%s)) ",
			accept_opts, deny_opts);
		goto write_query;
	}
	if (accept_opts && !deny_opts) {
		where_clause = make_message("and (%s) ", accept_opts);
		goto write_query;
	}
	if (!accept_opts && deny_opts) {
		where_clause = make_message("and not (%s) ", deny_opts);
		goto write_query;
	}
	where_clause = para_strdup("");
write_query:
	query = make_message("%s %s %s", select_clause, where_clause, order);
	free(order);
	free(select_clause);
	free(where_clause);
out:
	if (accept_opts)
		free(accept_opts);
	if (deny_opts)
		free(deny_opts);
	if (result)
		mysql_free_result(result);
	return query;
}



/*
 * This is called from server and from some commands. Name must not be NULL
 * Never returns NULL.
 */
static char *get_selector_info(char *name)
{
	char *meta, *atts, *info, *dir, *query, *stream;
	void *result = NULL;
	MYSQL_ROW row = NULL;

	if (!name)
		return para_strdup("(none)");
	stream = get_current_stream();
	meta = get_meta(name, 1);
	atts = get_atts(name, 0);
	dir = get_dir(name);
	/* get score */
	query = get_query(stream, name, 0); /* FIXME: pass stream == NULL instead? */
	if (!query)
		goto write;
	result = get_result(query);
	free(query);
	if (result)
		row = mysql_fetch_row(result);
write:
	info = make_message("dbinfo1:dir: %s\n"
		"dbinfo2:stream: %s, %s, score: %s\n"
		"dbinfo3:%s\n",
		dir? dir : "(not contained in table)",
		stream, meta, 
		(result && row && row[0])? row[0] : "(no score)",
		atts);
	free(dir);
	free(meta);
	free(atts);
	free(stream);
	if (result)
		mysql_free_result(result);
	return info;
}

/* might return NULL */
static char *get_current_audio_file(void)
{
	char *name;
	mmd_lock();
	name = para_basename(mmd->filename);
	mmd_unlock();
	return name;
}

/* If called as child, mmd_lock must be held */
static void update_mmd(char *info)
{
	PARA_DEBUG_LOG("%s", "updating shared memory area\n");
	strncpy(mmd->selector_info, info, MMD_INFO_SIZE - 1);
	mmd->selector_info[MMD_INFO_SIZE - 1] = '\0';
}

static void refresh_selector_info(void)
{
	char *name = get_current_audio_file();
	char *info;

	if (!name)
		return;
	info = get_selector_info(name);
	free(name);
	mmd_lock();
	update_mmd(info);
	mmd_unlock();
	free(info);
}

/* list attributes / print database info */
static int com_la_info(int fd, int argc, char *argv[])
{
	char *name = NULL, *meta = NULL, *atts = NULL, *dir = NULL;
	int ret, la = strcmp(argv[0], "info");

	if (argc < 2) {
		ret = -E_GET_AUDIO_FILE;
		if (!(name = get_current_audio_file()))
			goto out;
		ret = send_va_buffer(fd, "%s\n", name);
		if (ret < 0)
			goto out;
	} else {
		ret = -E_ESCAPE;
		if (!(name = escaped_basename(argv[1])))
			goto out;
	}
	meta = get_meta(name, 1);
	atts = get_atts(name, 0);
	dir = get_dir(name);
	if (la)
		ret = send_va_buffer(fd, "%s\n", atts);
	else
		ret = send_va_buffer(fd, "dir: %s\n" "%s\n" "attributes: %s\n",
			dir? dir : "(not contained in table)", meta, atts);
out:
	free(meta);
	free(atts);
	free(dir);
	free(name);
	return ret;
}

/* list attributes */
int com_la(int fd, int argc, char *argv[])
{
	return com_la_info(fd, argc, argv);
}

/* print database info */
int com_info(int fd, int argc, char *argv[])
{
	return com_la_info(fd, argc, argv);
}

static int change_stream(const char *stream)
{
	char *query;
	int ret;
	query = make_message("update streams set def='%s' "
		"where name = 'current_stream'", stream);
	ret = real_query(query);
	free(query);
	return ret;
}

static int get_pic_id_by_name(char *name)
{
	char *q, *ebn;
	void *result = NULL;
	int ret;
	MYSQL_ROW row;

	if (!(ebn = escaped_basename(name)))
		return -E_ESCAPE;
	q = make_message("select pic_id from data where name = '%s'", ebn);
	free(ebn);
	result = get_result(q);
	free(q);
	if (!result)
		return -E_NORESULT;
	row = mysql_fetch_row(result);
	ret = -E_NOROW;
	if (row && row[0])
		ret = atoi(row[0]);
	mysql_free_result(result);
	return ret;
}

static int remove_entry(const char *name)
{
	char *q, *ebn = escaped_basename(name);
	int ret = -E_ESCAPE;

	if (!ebn)
		goto out;
	q = make_message("delete from data where name = '%s'", ebn);
	real_query(q); /* ignore errors */
	free(q);
	q = make_message("delete from dir where name = '%s'", ebn);
	real_query(q); /* ignore errors */
	free(q);
	ret = 1;
out:
	free(ebn);
	return ret;
}

static int add_entry(const char *name)
{
	char *q, *dn, *ebn = NULL, *edn = NULL;
	int ret;

	if (!name || !*name)
		return -E_MYSQL_SYNTAX;
	ebn = escaped_basename(name);
	if (!ebn)
		return -E_ESCAPE;
	ret = -E_MYSQL_SYNTAX;
	dn = para_dirname(name);
	if (!dn)
		goto out;
	ret = -E_ESCAPE;
	edn = escape_str(dn);
	free(dn);
	if (!edn || !*edn)
		goto out;
	q = make_message("insert into data (name, pic_id) values "
			"('%s', '%s')", ebn, "1");
	ret = real_query(q);
//	ret = 1; PARA_DEBUG_LOG("q: %s\n", q);
	free(q);
	if (ret < 0)
		goto out;
	q = make_message("insert into dir (name, dir) values "
			"('%s', '%s')", ebn, edn);
//	ret = 1; PARA_DEBUG_LOG("q: %s\n", q);
	ret = real_query(q);
	free(q);
out:
	if (ebn)
		free(ebn);
	if (edn)
		free(edn);
	return ret;
}

/*
 * remove/add entries
 */
static int com_rm_ne(__a_unused int fd, int argc, char *argv[])
{
	int ne = !strcmp(argv[0], "ne");
	int i, ret;
	if (argc < 2)
		return -E_MYSQL_SYNTAX;
	for (i = 1; i < argc; i++) {
		ret = remove_entry(argv[i]);
		if (ret < 0)
			return ret;
		if (!ne)
			continue;
		ret = add_entry(argv[i]);
		if (ret < 0)
			return ret;
	}
	return 1;
}

/*
 * rm
 */
int com_rm(int fd, int argc, char *argv[])
{
	return com_rm_ne(fd, argc, argv);
}

/*
 * ne
 */
int com_ne(int fd, int argc, char *argv[])
{
	return com_ne(fd, argc, argv);
}

/*
 * mv: rename entry
 */
int com_mv(__a_unused int fd, int argc, char *argv[])
{
	char *q, *dn, *ebn1 = NULL, *ebn2 = NULL, *edn = NULL;
	int ret;

	if (argc != 3)
		return -E_MYSQL_SYNTAX;
	ret = -E_ESCAPE;
	ebn1 = escaped_basename(argv[1]);
	ebn2 = escaped_basename(argv[2]);
	if (!ebn1 || !ebn2 || !*ebn1 || !*ebn2)
		goto out;
	ret = -E_MYSQL_SYNTAX;
	if (!strcmp(ebn1, ebn2))
		goto update_dir;
	remove_entry(argv[2]); /* no need to escape, ignore error */
	q = make_message("update data set name = '%s' where name = '%s'",
		ebn2, ebn1);
	ret = real_query(q);
	free(q);
	if (ret < 0)
		goto out;
	ret = -E_AUDIO_FILE;
	if (!mysql_affected_rows(mysql_ptr))
		goto out;
	q = make_message("update dir set name = '%s' where name = '%s'",
		ebn2, ebn1);
	ret = real_query(q);
	free(q);
	if (ret < 0)
		goto out;
update_dir:
	ret = 1;
	dn = para_dirname(argv[2]);
	if (!dn)
		goto out;
	ret = -E_ESCAPE;
	edn = escape_str(dn);
	free(dn);
	if (!edn)
		goto out;
	ret = 1;
	if (!*edn)
		goto out;
	q = make_message("update dir set dir = '%s' where name = '%s'",
		edn, ebn2);
	ret = real_query(q);
	free(q);
out:
	free(edn);
	free(ebn1);
	free(ebn2);
	return ret;
}

/*
 * set field
 */
static int com_set(__a_unused int fd, int argc, char *argv[])
{
	char *q, *ebn;
	long unsigned id;
	int i, ret;
	const char *field = strcmp(argv[0], "picass")? "numplayed" : "pic_id";

	if (argc < 3)
		return -E_MYSQL_SYNTAX;
	id = atol(argv[1]);
	for (i = 2; i < argc; i++) {
		ebn = escaped_basename(argv[i]);
		if (!ebn)
			return -E_ESCAPE;
		q = make_message("update data set %s = %lu "
				"where name = '%s'", field, id, ebn);
		free(ebn);
		ret = real_query(q);
		free(q);
		if (ret < 0)
			return ret;
	}
	return 1;
}

/*
 * snp: set numplayed
 */
int com_picass(int fd, int argc, char *argv[])
{
	return com_set(fd, argc, argv);
}

/*
 * snp: set numplayed
 */
int com_snp(int fd, int argc, char *argv[])
{
	int ret = com_set(fd, argc, argv);
	if (ret >= 0)
		refresh_selector_info();
	return ret;
}

/*
 * picch: change entry's name in pics table
 */
int com_picch(__a_unused int fd, int argc, char *argv[])
{
	int ret;
	long unsigned id;
	char *q, *tmp;

	if (argc != 3)
		return -E_MYSQL_SYNTAX;
	id = atol(argv[1]);
	ret = -E_ESCAPE;
	tmp = escape_str(argv[2]);
	if (!tmp)
		return -E_ESCAPE;
	q = make_message("update pics set name = '%s' where id = %lu", tmp, id);
	free(tmp);
	ret = real_query(q);
	free(q);
	return ret;
}

/*
 * piclist: print list of pics in db
 */
int com_piclist(__a_unused int fd, int argc, __a_unused char *argv[])
{
	void *result = NULL;
	MYSQL_ROW row;
	unsigned long *length;
	int ret;

	if (argc != 1)
		return -E_MYSQL_SYNTAX;
	result = get_result("select id,name,pic from pics order by id");
	if (!result)
		return -E_NORESULT;
	while ((row = mysql_fetch_row(result))) {
		length = mysql_fetch_lengths(result);
		if (!row || !row[0] || !row[1] || !row[2])
			continue;
		ret = send_va_buffer(fd, "%s\t%lu\t%s\n", row[0], length[2], row[1]);
		if (ret < 0)
			goto out;
	}
	ret = 1;
out:
	mysql_free_result(result);
	return ret;
}

/*
 * picdel: delete picture from database
 */
int com_picdel(int fd, int argc, char *argv[])
{
	char *q;
	long unsigned id;
	my_ulonglong aff;
	int i, ret;

	if (argc < 2)
		return -E_MYSQL_SYNTAX;
	for (i = 1; i < argc; i++) {
		id = atol(argv[i]);
		q = make_message("delete from pics where id = %lu", id);
		ret = real_query(q);
		free(q);
		if (ret < 0)
			return ret;
		aff = mysql_affected_rows(mysql_ptr);
		if (!aff) {
			ret = send_va_buffer(fd, "No such id: %lu\n", id);
			if (ret < 0)
				return ret;
			continue;
		}
		q = make_message("update data set pic_id = 1 where pic_id = %lu", id);
		ret = real_query(q);
		free(q);
	}
	return 1;
}
/*
 * pic: get picture by name or by number
 */
int com_pic(int fd, int argc, char *argv[])
{
	void *result = NULL;
	MYSQL_ROW row;
	unsigned long *length, id;
	int ret;
	char *q, *name = NULL;

	if (argc < 2) {
		ret = -E_GET_AUDIO_FILE;
		name = get_current_audio_file();
	} else {
		ret = -E_ESCAPE;
		name = escaped_basename(argv[1]);
	}
	if (!name)
		return ret;
	if (*name == '#')
		id = atoi(name + 1);
	else
		id = get_pic_id_by_name(name);
	free(name);
	if (id <= 0)
		return id;
	q = make_message("select pic from pics where id = '%lu'", id);
	result = get_result(q);
	free(q);
	if (!result)
		return -E_NORESULT;
	row = mysql_fetch_row(result);
	ret = -E_NOROW;
	if (!row || !row[0])
		goto out;
	length = mysql_fetch_lengths(result);
	ret = send_bin_buffer(fd, row[0], *length);
out:
	mysql_free_result(result);
	return ret;
}

/* strdel */
int com_strdel(__a_unused int fd, int argc, char *argv[])
{
	char *q, *tmp;
	int ret;

	if (argc < 2)
		return -E_MYSQL_SYNTAX;
	tmp = escape_str(argv[1]);
	if (!tmp)
		return -E_ESCAPE;
	q = make_message("delete from streams where name='%s'", tmp);
	free(tmp);
	ret = real_query(q);
	free(q);
	return ret;
}

/*
 * ls
 */
int com_ls(int fd, int argc, char *argv[])
{
	char *q;
	void *result;
	int ret;
	my_ulonglong num_rows, top = 0, left = 0, right = 0;

	if (argc > 2)
		return -E_MYSQL_SYNTAX;
	if (argc > 1) {
		char *tmp = escape_str(argv[1]);
		if (!tmp)
			return -E_ESCAPE;
		q = make_message("select name from data where name like '%s'",
			tmp);
		free(tmp);
	} else
		q = para_strdup("select name from data");
	result = get_result(q);
	free(q);
	if (!result)
		return -E_NORESULT;
	num_rows = mysql_num_rows(result);
	ret = 1;
	if (num_rows)
		ret = print_results(fd, result, top, left, num_rows - 1, right);
	mysql_free_result(result);
	return ret;
}

/*
 * summary
 */
int com_summary(__a_unused int fd, int argc, __a_unused char *argv[])
{
	MYSQL_ROW row;
	MYSQL_ROW row2;
	void *result;
	void *result2 = NULL;
	const char *fmt = "select count(name) from data where %s='1'";
	int ret = -E_NORESULT;

	if (argc != 1)
		return -E_MYSQL_SYNTAX;
	result = get_all_attributes();
	if (!result)
		goto out;
	while ((row = mysql_fetch_row(result))) {
		char *buf;

		ret = -E_NOROW;
		if (!row[0])
			goto out;
		ret = -E_NORESULT;
		buf = make_message(fmt, row[0]);
		result2 = get_result(buf);
		free(buf);
		if (!result2)
			goto out;
		ret = -E_NOROW;
		row2 = mysql_fetch_row(result2);
		if (!row2 || !row2[0])
			goto out;
		ret = send_va_buffer(fd, "%s\t%s\n", row[0], row2[0]);
		if (ret < 0)
			goto out;
	}
	ret = 1;
out:
	if (result2)
		mysql_free_result(result2);
	if (result)
		mysql_free_result(result);
	return ret;
}

static int get_numplayed(char *name)
{
	void *result;
	MYSQL_ROW row;
	const char *fmt = "select numplayed from data where name = '%s'";
	char *buf = make_message(fmt, name);
	int ret = -E_NORESULT;

	result = get_result(buf);
	free(buf);
	if (!result)
		goto out;
	ret = -E_NOROW;
	row = mysql_fetch_row(result);
	if (!row || !row[0])
		goto out;
	ret = atoi(row[0]);
out:
	if (result)
		mysql_free_result(result);
	return ret;
}

static int update_audio_file(char *name)
{
	int ret;
	const char *fmt1 = "update data set lastplayed = now() where name = '%s'";
	const char *fmt2 = "update data set numplayed = %i where name = '%s'";
	char *q;
	char *ebn = escaped_basename(name);

	ret = -E_ESCAPE;
	if (!ebn)
		goto out;
	q = make_message(fmt1, ebn);
	ret = real_query(q);
	free(q);
	if (ret < 0)
		goto out;
	ret = get_numplayed(ebn);
	if (ret < 0)
		goto out;
	q = make_message(fmt2, ret + 1, ebn);
	ret = real_query(q);
	free(q);
out:
	free(ebn);
	return ret;
}

static void update_audio_file_server_handler(char *name)
{
	char *info;
	info = get_selector_info(name);
	update_mmd(info);
	free(info);
	update_audio_file(name);
}

int com_us(__a_unused int fd, int argc, char *argv[])
{
	char *tmp;
	int ret;

	if (argc != 2)
		return -E_MYSQL_SYNTAX;
	tmp = escape_str(argv[1]);
	if (!tmp)
		return -E_ESCAPE;
	ret = update_audio_file(argv[1]);
	free(tmp);
	if (ret >= 0)
		refresh_selector_info();
	return ret;
}

/* select previous / next stream */
static int com_ps_ns(__a_unused int fd, int argc, char *argv[])
{
	char *query, *stream = get_current_stream();
	void *result = get_result("select name from streams");
	MYSQL_ROW row;
	int ret;
	my_ulonglong num_rows, match, i;

	ret = -E_MYSQL_SYNTAX;
	if (argc != 1)
		goto out;
	ret = -E_NORESULT;
	if (!result)
		goto out;
	num_rows = mysql_num_rows(result);
	ret = -E_EMPTY_RESULT;
	if (num_rows < 2)
		goto out;
	ret = -E_NOROW;
	for (i = 0; i < num_rows; i++) {
		row = mysql_fetch_row(result);
		if (!row || !row[0])
			goto out;
		if (!strcmp(row[0], "current_stream"))
			continue;
		if (!strcmp(row[0], stream))
			break;
	}
	ret = -E_NO_STREAM;
	if (i == num_rows)
		goto out;
	match = i;
	if (!strcmp(argv[0], "ps"))
		i = match > 0? match - 1 : num_rows - 1;
	else
		i = match < num_rows - 1? match + 1 : 0;
	ret = -E_NOROW;
	mysql_data_seek(result, i);
	row = mysql_fetch_row(result);
	if (!row || !row[0])
		goto out;
	if (!strcmp(row[0], "current_stream")) {
		if (!strcmp(argv[0], "ps")) {
			i = match < 2? match + num_rows - 2 : match - 2;
		} else {
			i = match + 2;
			i = i > num_rows - 1? i - num_rows : i;
		}
		mysql_data_seek(result, i);
		row = mysql_fetch_row(result);
		if (!row || !row[0])
			goto out;
	}
	query = make_message("update streams set def='%s' where name = "
		"'current_stream'", row[0]);
	ret = real_query(query);
	free(query);
	refresh_selector_info();
out:
	free(stream);
	if (result)
		mysql_free_result(result);
	return ret;
}

/* select previous stream */
int com_ps(int fd, int argc, char *argv[])
{
	return com_ps_ns(fd, argc, argv);
}

/* select next stream */
int com_ns(int fd, int argc, char *argv[])
{
	return com_ps_ns(fd, argc, argv);
}

/* streams */
int com_streams(int fd, int argc, __a_unused char *argv[])
{
	unsigned int num_rows;
	int i, ret = -E_NORESULT;
	void *result;
	MYSQL_ROW row;

	if (argc > 1 && strcmp(argv[1], "current_stream"))
		return -E_MYSQL_SYNTAX;
	if (argc > 1) {
		char *cs = get_current_stream();
		ret = send_va_buffer(fd, "%s\n", cs);
		free(cs);
		return ret;
	}
	result = get_result("select name from streams");
	if (!result)
		goto out;
	num_rows = mysql_num_rows(result);
	ret = 1;
	if (!num_rows)
		goto out;
	ret = -E_NOROW;
	for (i = 0; i < num_rows; i++) {
		row = mysql_fetch_row(result);
		if (!row || !row[0])
			goto out;
		if (strcmp(row[0], "current_stream"))
			send_va_buffer(fd, "%s\n", row[0]);
	}
	ret = 1;
out:
	if (result)
		mysql_free_result(result);
	return ret;
}

/* query stream definition */
int com_strq(int fd, int argc, char *argv[])
{
	MYSQL_ROW row;
	char *query, *name;
	void *result;
	int ret;

	if (argc < 2) {
		ret = -E_GET_STREAM;
		name = get_current_stream();
	} else {
		ret = -E_ESCAPE;
		name = escaped_basename(argv[1]);
	}
	if (!name)
		return ret;
	ret = -E_NORESULT;
	query = make_message("select def from streams where name='%s'", name);
	free(name);
	result = get_result(query);
	free(query);
	if (!result)
		goto out;
	ret = -E_NOROW;
	row = mysql_fetch_row(result);
	if (!row || !row[0])
		goto out;
	/* no '\n' needed */
	ret = send_buffer(fd, row[0]);
out:
	if (result)
		mysql_free_result(result);
	return ret;
}

/* change stream / change stream and play */
static int com_cs_csp(int fd, int argc, char *argv[])
{
	int ret, stream_change;
	char *query, *stream = NULL;
	char *old_stream = get_current_stream();
	int csp = !strcmp(argv[0], "csp");

	ret = -E_MYSQL_SYNTAX;
	if (argc > 2)
		goto out;
	if (argc == 1) {
		if (csp)
			goto out;
		ret = send_va_buffer(fd, "%s\n", old_stream);
		goto out;
	}
	ret = -E_GET_QUERY;
	/* test if stream is valid, no need to escape argv[1] */
	query = get_query(argv[1], NULL, 0);
	if (!query)
		goto out;
	free(query);
	/* stream is ok */
	stream = escape_str(argv[1]);
	if (!stream)
		goto out;
	stream_change = strcmp(stream, old_stream);
	if (stream_change) {
		ret = change_stream(stream);
		if (ret < 0)
			goto out;
		refresh_selector_info();
	}
	if (csp) {
		mmd_lock();
		mmd->new_vss_status_flags |= VSS_PLAYING;
		if (stream_change)
			mmd->new_vss_status_flags |= VSS_NEXT;
		mmd_unlock();
	}
	ret = 1;
out:
	free(old_stream);
	free(stream);
	return ret;
}

/* change stream */
int com_cs(int fd, int argc, char *argv[])
{
	return com_cs_csp(fd, argc, argv);
}

/* change stream and play */
int com_csp(int fd, int argc, char *argv[])
{
	return com_cs_csp(fd, argc, argv);
}

/* score list / skip */
static int com_sl_skip(int fd, int argc, char *argv[])
{
	void *result = NULL;
	MYSQL_ROW row;
	int ret, i, skip = !strcmp(argv[0], "skip");
	char *query, *stream, *tmp;
	unsigned int num_rows, num;

	if (argc < 2)
		return -E_MYSQL_SYNTAX;
	num = atoi(argv[1]);
	if (!num)
		return -E_MYSQL_SYNTAX;
	if (argc == 2) {
		stream = get_current_stream();
		if (!stream)
			return -E_GET_STREAM;
	} else {
		stream = escape_str(argv[2]);
		if (!stream)
			return -E_ESCAPE;
	}
	tmp = get_query(stream, NULL, 0);
	free(stream);
	if (!tmp)
		return -E_GET_QUERY;
	query = make_message("%s limit %d", tmp, num);
	free(tmp);
	ret = -E_NORESULT;
	result = get_result(query);
	free(query);
	if (!result)
		goto out;
	ret = -E_EMPTY_RESULT;
	num_rows = mysql_num_rows(result);
	if (!num_rows)
		goto out;
	for (i = 0; i < num_rows && i < num; i++) {
		row = mysql_fetch_row(result);
		if (skip) {
			send_va_buffer(fd, "Skipping %s\n", row[0]);
			update_audio_file(row[0]);
		} else
			send_va_buffer(fd, "%s\n", row[0]? row[0]: "BUG");
	}
	ret = 1;
out:
	if (result)
		mysql_free_result(result);
	return ret;
}

/* score list */
int com_sl(int fd, int argc, char *argv[])
{
	return com_sl_skip(fd, argc, argv);
}

/* skip */
int com_skip(int fd, int argc, char *argv[])
{
	return com_sl_skip(fd, argc, argv);
}

/*
 * update attributes of name
 */
static int update_atts(int fd, char *name, char *atts)
{
	int ret;
	char *ebn, *q, *old, *new = NULL;

	if (!mysql_ptr)
		return -E_NOTCONN;
	ebn = escaped_basename(name);
	if (!ebn)
		return -E_ESCAPE;
	q = make_message("update data set %s where name = '%s'", atts, ebn);
	old = get_atts(ebn, 0);
	send_va_buffer(fd, "old: %s\n", old);
	free(old);
	ret = real_query(q);
	free(q);
	if (ret < 0)
		goto out;
	new = get_atts(ebn, 0);
	ret = send_va_buffer(fd, "new: %s\n", new);
	free(new);
out:
	free(ebn);
	return ret;
}

/*
 * set attributes
 */
int com_sa(int fd, int argc, char *argv[])
{
	int i, ret;
	char *atts = NULL, *name;

	if (argc < 2)
		return -E_MYSQL_SYNTAX;
	for (i = 1; i < argc; i++) {
		int unset = 0;
		char *esc, *tmp, *p =argv[i];
		int len = strlen(p);

		if (!len)
			continue;
		switch (p[len - 1]) {
			case '+':
				unset = 0;
				break;
			case '-':
				unset = 1;
				break;
			default:
				goto no_more_atts;
		}
		p[len - 1] = '\0';
		esc = escape_str(p);
		if (!esc)
			return -E_ESCAPE;
		tmp = make_message("%s%s='%s'", atts? "," : "", esc,
			unset? "0" : "1");
		free(esc);
		atts = para_strcat(atts, tmp);
		free(tmp);
	}
no_more_atts:
	if (!atts)
		return -E_NOATTS;
	if (i >= argc) { /* no name given, use current af */
		ret = -E_GET_AUDIO_FILE;
		if (!(name = get_current_audio_file()))
			goto out;
		ret = update_atts(fd, name, atts);
		free(name);
	} else {
		ret = 1;
		for (; argv[i] && ret >= 0; i++)
			ret = update_atts(fd, argv[i], atts);
	}
	refresh_selector_info();
out:
	free(atts);
	return ret;
}

/*
 * copy attributes
 */
int com_cam(int fd, int argc, char *argv[])
{
	char *name = NULL, *meta = NULL, *atts = NULL;
	int i, ret;

	if (argc < 3)
		return -E_MYSQL_SYNTAX;
	if (!(name = escaped_basename(argv[1])))
		return -E_ESCAPE;
	ret = -E_NOATTS;
	if (!(atts = get_atts(name, 1)))
		goto out;
	ret = -E_META;
	if (!(meta = get_meta(name, 0)))
		goto out;
	for (i = 2; i < argc; i++) {
		char *ebn, *q;
		ret = -E_ESCAPE;
		if (!(ebn = escaped_basename(argv[i])))
			goto out;
		ret = send_va_buffer(fd, "updating %s\n", ebn);
		if (ret < 0) {
			free(ebn);
			goto out;
		}
		q = make_message("update data set %s where name = '%s'",
			meta, ebn);
		if ((ret = update_atts(fd, ebn, atts)) >= 0)
			ret = real_query(q);
		free(ebn);
		free(q);
		if (ret < 0)
			goto out;
	}
	ret = 1;
out:
	if (name)
		free(name);
	if (meta)
		free(meta);
	if (atts)
		free(atts);
	return ret;
}

/*
 * verify / clean
 */
static int com_vrfy_clean(int fd, int argc, __a_unused char *argv[])
{
	char *query;
	int ret, vrfy_mode = strcmp(argv[0], "clean");
	void *result = NULL;
	MYSQL_ROW row;
	char *escaped_name;
	my_ulonglong num_rows, top = 0, left = 0, right = 0;

	if (argc != 1)
		return -E_MYSQL_SYNTAX;
	ret = -E_NORESULT;
	result = get_result("select data.name from data left join dir on "
		"dir.name = data.name where dir.name is NULL");
	if (!result)
		goto out;
	num_rows = mysql_num_rows(result);
	if (!num_rows) {
		ret = send_buffer(fd, "No invalid entries\n");
		goto out;
	}
	if (vrfy_mode) {
		send_va_buffer(fd, "found %lli invalid entr%s\n", num_rows,
			num_rows == 1? "y" : "ies");
		ret = print_results(fd, result, top, left, num_rows - 1, right);
		goto out;
	}
	while ((row = mysql_fetch_row(result))) {
		ret = -E_NOROW;
		if (!row[0])
			goto out;
		ret = -E_ESCAPE;
		escaped_name = escape_str(row[0]);
		if (!escaped_name)
			goto out;
		send_va_buffer(fd, "deleting %s\n", escaped_name);
		query = make_message("delete from data where name = '%s'",
			escaped_name);
		ret = real_query(query);
		free(query);
		if (ret < 0)
			goto out;
	}

out:
	if (result)
		mysql_free_result(result);
	return ret;
}

/*
 * verify
 */
int com_vrfy(int fd, int argc, char **argv)
{
	return com_vrfy_clean(fd, argc, argv);
}

/*
 * clean
 */
int com_clean(int fd, int argc, char **argv)
{
	return com_vrfy_clean(fd, argc, argv);
}

static FILE *out_file;

static int mysql_write_tmp_file(const char *dir, const char *name)
{
	int ret = -E_TMPFILE;
	char *msg = make_message("%s\t%s\n", dir, name);
	if (fputs(msg, out_file) != EOF)
		ret = 1;
	free(msg);
	return ret;
}

/*
 * update database
 */
int com_upd(int fd, int argc, __a_unused char *argv[])
{
	char *tempname = NULL, *query = NULL;
	int ret, out_fd = -1, num = 0;
	void *result = NULL;
	unsigned int num_rows;
	MYSQL_ROW row;

	if (argc != 1)
		return -E_MYSQL_SYNTAX;
	out_file = NULL;
	tempname = para_strdup("/tmp/mysql.tmp.XXXXXX");
	ret = para_mkstemp(tempname, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
	if (ret < 0)
		goto out;
	out_fd = ret;
	out_file = fdopen(out_fd, "w");
	if (!out_file) {
		close(out_fd);
		goto out;
	}
	if (find_audio_files(conf.mysql_audio_file_dir_arg, mysql_write_tmp_file) < 0)
		goto out;
	num = ftell(out_file);
	/*
	 * we have to make sure the file hit the disk before we call
	 * real_query
	 */
	fclose(out_file);
	out_file = NULL;
	PARA_DEBUG_LOG("wrote tempfile %s (%d bytes)\n", tempname, num);
	if (!num)
		goto out;
	if ((ret = real_query("delete from dir")) < 0)
		goto out;
	query = make_message("load data infile '%s' ignore into table dir "
		"fields terminated by '\t' lines terminated by '\n' "
		"(dir, name)", tempname);
	ret = real_query(query);
	free(query);
	if (ret < 0)
		goto out;
	result = get_result("select dir.name from dir left join data on "
		"data.name = dir.name where data.name is NULL");
	ret = -E_NORESULT;
	if (!result)
		goto out;
	num_rows = mysql_num_rows(result);
	if (!num_rows) {
		ret = send_buffer(fd, "no new entries\n");
		goto out;
	}
	while ((row = mysql_fetch_row(result))) {
		char *erow;
		ret = -E_NOROW;
		if (!row[0])
			goto out;
		send_va_buffer(fd, "new entry: %s\n", row[0]);
		erow = escape_str(row[0]);
		if (!erow)
			goto out;
		query = make_message("insert into data (name, pic_id) values "
			"('%s','%s')", erow, "1");
		free(erow);
		ret = real_query(query);
		free(query);
		if (ret < 0)
			goto out;
	}
	ret = 1;
out:
	if (out_fd >= 0)
		unlink(tempname);
	free(tempname);
	if (out_file)
		fclose(out_file);
	if (result)
		mysql_free_result(result);
	return ret;
}

static char **server_get_audio_file_list(unsigned int num)
{
	char **list = para_malloc((num + 1) * sizeof(char *));
	char *tmp, *query, *stream = get_current_stream();
	void *result = NULL;
	unsigned int num_rows;
	int i = 0;
	MYSQL_ROW row;

	tmp = get_query(stream, NULL, 1);
	free(stream);
	if (!tmp)
		goto err_out;
	query = make_message("%s limit %d", tmp, num);
	free(tmp);
	result = get_result(query);
	free(query);
	if (!result)
		goto err_out;
	num_rows = mysql_num_rows(result);
	if (!num_rows)
		goto err_out;
	for (i = 0; i < num_rows && i < num; i++) {
		row = mysql_fetch_row(result);
		if (!row || !row[0])
			goto err_out;
		list[i] = para_strdup(row[0]);
	}
	list[i] = NULL;
	goto success;
err_out:
	while (i > 0) {
		i--;
		free(list[i]);
	}
	free(list);
	list = NULL;
success:
	if (result)
		mysql_free_result(result);
	return list;
}

/*
 * connect to mysql server, return mysql pointer on success, -E_NOTCONN
 * on errors. Called from parent on startup and also from com_cdb().
 */
static int init_mysql_server(void)
{
	char *u = conf.mysql_user_arg? conf.mysql_user_arg : para_logname();
	unsigned int port;

	mysql_ptr = mysql_init(NULL);
	if (!mysql_ptr) {
		PARA_CRIT_LOG("%s", "mysql init error\n");
		return -E_NOTCONN;
	}
	if (conf.mysql_port_arg < 0)
		return -E_MYSQL_SYNTAX;
	port = conf.mysql_port_arg;
	PARA_DEBUG_LOG("connecting: %s@%s:%d\n", u, conf.mysql_host_arg, port);
	if (!conf.mysql_user_arg)
		free(u);
	/*
         * If host is NULL a connection to the local host is assumed,
         * If user is NULL, the current user is assumed
	 */
	if (!(mysql_ptr = mysql_real_connect(mysql_ptr,
			conf.mysql_host_arg,
			conf.mysql_user_arg,
			conf.mysql_passwd_arg,
			conf.mysql_database_arg,
			port, NULL, 0))) {
		PARA_CRIT_LOG("%s", "connect error\n");
		return -E_NOTCONN;
	}
	PARA_INFO_LOG("%s", "success\n");
	return 1;
}

/* mmd lock must be held */
static void write_msg2mmd(int success)
{
	sprintf(mmd->selector_info, "dbinfo1:%s\ndbinfo2:mysql-%s\ndbinfo3:\n",
		success < 0? PARA_STRERROR(-success) :
			"successfully connected to mysql server",
		success < 0? "" : mysql_get_server_info(mysql_ptr));
}

/* create database */
int com_cdb(int fd, int argc, char *argv[])
{
	char *query;
	int ret;

	if (mysql_ptr) {
		PARA_INFO_LOG("%s", "closing database\n");
		mysql_close(mysql_ptr);
	}
	/* dont use any database */
	conf.mysql_database_arg = NULL; /* leak? */
	ret = -E_MYSQL_INIT;
	if (init_mysql_server() < 0 || !mysql_ptr)
		goto out;
	if (argc < 2)
		conf.mysql_database_arg = para_strdup("paraslash");
	else {
		ret = -E_ESCAPE;
		conf.mysql_database_arg = escape_str(argv[1]);
		if (!conf.mysql_database_arg)
			goto out;
	}
	query = make_message("create database %s", conf.mysql_database_arg);
	ret = real_query(query);
	free(query);
	if (ret < 0)
		goto out;
	/* reconnect with database just created */
	mysql_close(mysql_ptr);
	ret = -E_MYSQL_INIT;
	if (init_mysql_server() < 0 || !mysql_ptr)
		goto out;
	mmd_lock();
	write_msg2mmd(1);
	mmd_unlock();
	ret = -E_QFAILED;
	if (real_query("create table data (name varchar(255) binary not null "
			"primary key, "
			"lastplayed datetime not null default "
				"'1970-01-01', "
			"numplayed int not null default 0, "
			"pic_id bigint unsigned not null default 1)") < 0)
		goto out;
	if (real_query("create table dir (name varchar(255) binary not null "
			"primary key, dir varchar(255) default null)") < 0)
		goto out;
	if (real_query("create table pics ("
			"id bigint(20) unsigned not null primary key "
			"auto_increment, "
			"name varchar(255) binary not null, "
			"pic mediumblob not null)") < 0)
		goto out;
	if (real_query("create table streams ("
			"name varchar(255) binary not null primary key, "
			"def blob not null)") < 0)
		goto out;
	if (real_query("insert into streams (name, def) values "
			"('current_stream', '(none)')") < 0)
		goto out;
	ret = send_va_buffer(fd, "successfully created database %s\n",
		conf.mysql_database_arg);
out:
	return ret;
}

static void shutdown_connection(void)
{
	int ret;

	if (mysql_ptr) {
		PARA_NOTICE_LOG("%s", "shutting down mysql connection\n");
		mysql_close(mysql_ptr);
		mysql_ptr = NULL;
	}
	if (mysql_lock) {
		ret = mutex_destroy(mysql_lock);
		if (ret < 0)
			PARA_ERROR_LOG("%s\n", PARA_STRERROR(-ret));
		mysql_lock = 0;
	}
}

/**
 * the init function of the mysql-based audio file selector
 *
 * \param db pointer to the struct to initialize
 *
 * Check the command line options and initialize all function pointers of \a
 * db.  Connect to the mysql server and initialize the info string.
 *
 * \return This function returns success even if it could not connect
 * to the mysql server. This is because the connect is expected to fail
 * if there the paraslash database is not yet created. This gives the
 * user a chance to send the "cdb" to create the database.
 *
 * \sa struct audio_file_selector, misc_meta_data::selector_info,
 * random_selector.c
 */
int mysql_selector_init(struct audio_file_selector *db)
{
	int ret;

	if (!conf.mysql_passwd_given)
		return -E_NO_MYSQL_PASSWD;
	if (!conf.mysql_audio_file_dir_given)
		return -E_NO_AF_DIR;
	db->name = "mysql";
	db->cmd_list = mysql_selector_cmds;
	db->get_audio_file_list = server_get_audio_file_list;
	db->update_audio_file = update_audio_file_server_handler;
	db->shutdown = shutdown_connection;
	ret = mutex_new();
	if (ret < 0)
		return ret;
	mysql_lock = ret;
	ret = init_mysql_server();
	if (ret < 0)
		PARA_WARNING_LOG("%s\n", PARA_STRERROR(-ret));
	write_msg2mmd(ret);
	return 1;	/* return success even if connect failed to give the
			 * user the chance to exec com_cdb
			 */
}
